import ntplib
import time

def sync_time(ntp_server='pool.ntp.org'):

    c = ntplib.NTPClient()
    response = c.request(ntp_server)
    offset = response.offset

    original_timestamp = time.time
    def correct_time():
        return original_timestamp() + offset

    time.time = correct_time



#!/usr/bin/env python3
# from hashlib import sha256 as hash_alg
from hashlib import sha1 as hash_alg
import codecs
import secrets
import base64

block_size = hash_alg().block_size

opad = bytes((x ^ 0x5c) for x in range(2**8))
ipad = bytes((x ^ 0x36) for x in range(2**8))


# RFC 2104 compliant
def hmac(K, m):

    if len(K) > block_size:
        k = hash_alg(K).digest()
    else:
        k = K

    k = k.ljust(block_size, b'\0')

    o_key_pad = k.translate(opad) # xor(k, k.translate(opad))
    i_key_pad = k.translate(ipad) # xor(k, k.translate(ipad))

    inner = hash_alg(i_key_pad)
    outer = hash_alg(o_key_pad)

    inner.update(m)
    outer.update(inner.digest())

    return outer.digest()


def otp(secret, moving_factor):

    c = "%016x" % moving_factor
    c = codecs.decode(c, 'hex')

    r = hmac(secret, c)

    offset = r[-1] & 0xf
    code = (r[offset] & 0x7f) << 24
    code = code | ((r[offset + 1] & 0xff) << 16)
    code = code | ((r[offset + 2] & 0xff) << 8)
    code = code | ((r[offset + 3] & 0xff))

    res = code % 1e10

    return int(res)


def hotp(K, I):
    res = otp(K, I) % 1e6
    return int(res)


# TC = Time in seconds
def totp(K, TC, n=6):

    t = int(TC/30)
    res = otp(K, t) % (10**n)
    return int(res)


def int_2_str(i):

    a = (i >> 24) & 0xff
    b = (i >> 16)  & 0xff
    c = (i >> 8) & 0xff
    d = i & 0xff

    return "%c%c%c%c" % (a, b, c, d)


def gen_secret():
    # TODO Better randomness?
    return codecs.decode(base64.b32encode(secrets.token_bytes(32))).replace('=', '')


def bytes2hexs(b):
    return codecs.encode(b, "hex")


def str2hexs(s):
    res = codecs.encode(s)
    res = bytes2hexs(res)
    return res


# Método para utilizar secretos 2fa de webs
def web_secret_2_bytes(s):

    norm = s.replace(' ', '')
    norm += '=' * (len(norm) % 8)
    key = base64.b32decode(norm)

    return key



from .proc_worker import ProcWorker, Event, PortManagerEvent, TocTocPortsEvent
from . import totp
import time
import threading
import logging
import random

log = logging.getLogger(__name__)


# 0 <= val < 1000000 --> 6 digits (24 bytes)
# 1024 < port < 10240
# 3 <= n_ports <= 6
def gen_ports_from_pin(pin, n_ports):

    values = []

    aux = pin


    min_port = 2**10 # 1024
    max_port = 2**16 # 65536

    for i in range(n_ports):
        random.seed(aux)

        aux = random.randint(min_port, max_port)

        values.append(aux)

    return values

class PortList():

    def __init__(self, l):
        self._l = l
        self._actual = 0

    def actual(self):
        return self._l[self._actual]

    def next(self):

        if len(self._l) <= self._actual:
            # raise Last()
            return 0

        n = self._l[self._actual]
        self._actual += 1
        return n

    def prev(self):

        if len(0 <= self._actual):
            return -1

        n = self._l[self._actual]
        self._actual -= 1
        return n

    def get_values(self):
        return self._l

    def reset(self):
        self._actual = 0

class TocTocPorts():

    def __init__(self, secret, slot=30, n_ports=4, destination=[]):

        self._secret = secret
        self._slot = slot

        self._destination = destination

        if n_ports < 1:
            raise Exception("Error, at least %d ports needed" % 1)

        if n_ports > 10:
            raise Exception("Error, max ports: %d" % 10)

        self._n_ports = n_ports

        ports = self.get_all()

        self._p = ports['p']
        self._a = ports['a']
        self._n = ports['n']
        # time.sleep(ns)

    def get_slot(self):
        return self._slot

    def get_destination(self):
        return self._destination

    def next(self):
        t = int(time.time())
        remainder = t % self._slot
        return self._slot - remainder

    def last(self):
        t = int(time.time())
        remainder = t % self._slot

        return t - remainder

    def get_all(self):
        return {'p': self.get_prev(), 'a': self.get_actual(), 'n': self.get_next()}

    def get_prev(self):

        tcp = self.last() - self._slot
        valp = totp.totp(self._secret, tcp)

        portsp = gen_ports_from_pin(valp, self._n_ports)

        return PortList(portsp)

    def get_actual(self):

        tca = self.last()
        vala = totp.totp(self._secret, tca)
        portsa = gen_ports_from_pin(vala, self._n_ports)

        return PortList(portsa)

    def get_next(self):

        tcn = self.last() + self._slot
        valn = totp.totp(self._secret, tcn)
        portsn = gen_ports_from_pin(valn, self._n_ports)

        return PortList(portsn)

    def __str__(self):
        def row(values):
            return "| " + ("{:<15}| "*len(values)).format(*values) + "\n"

        res = ''
        banner = row(["N", "Prev", "Actu", "Next"]) # "N\tPrev\t\tActu\t\tNext\n"
        res += ("-" * (len(banner) - 2))
        res += "\n"
        res += (banner)
        res += ("-" * (len(banner) - 2))
        res += "\n"

        ports = self.get_all()
        p = ports['p'].get_values()
        a = ports['a'].get_values()
        n = ports['n'].get_values()

        for port in range(len(p)):
            res += (row((port, p[port], a[port], n[port])))
        res += ("-" * (len(banner) - 2))
        res += "\n"

        res += " [*] NEXT_SLOT: %ds\n" % self.next()
        return res


class TocTocPortsWorker(ProcWorker):

    def __init__(self, i_q, o_q, ttp):
        super(TocTocPortsWorker, self).__init__(i_q, o_q)
        self._ttp = ttp
        threading.Thread(target=self.work).start()

    # TODO Cerrar este thread
    def work(self):

        while not self._end_evt.is_set():
            # TODO Tal vez no desde aquí, pero hay que lanzar un evento con los puertos reservados
            self._o.put(Event(TocTocPortsEvent.NEW_SLOT, {'port_list': self._ttp.get_actual()}))
            next_t = self._ttp.next()
            log.debug("Next slot in %ds" % next_t)
            self._end_evt.wait(next_t)

        log.debug("Fin del thread de TocTocPorts")

    def process_evt(self, evt):

        super(TocTocPortsWorker, self).process_evt(evt)

        if evt.get_id() == PortManagerEvent.LAST_PORT:
            self._o.put(Event(TocTocPortsEvent.LAST_PORT, {'ports': self._ttp.get_destination(), 'address': evt.get_value()['address']}))



#Packet sniffer in python for Linux
#Sniffs only incoming TCP packet

import socket, sys
from struct import *

#create an INET, STREAMing socket
try:
	s = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_TCP)
except socket.error:
# except msg:
	print( 'Socket could not be created. Error Code : ' + str(msg[0]) + ' Message ' + msg[1])
	sys.exit()

# receive a packet
while True:
	packet = s.recvfrom(65565)

	#packet string from tuple
	packet = packet[0]

	#take first 20 characters for the ip header
	ip_header = packet[0:20]
	# print(ip_header)

	#now unpack them :)
	iph = unpack('!BBHHHBBH4s4s' , ip_header)

	version_ihl = iph[0]
	version = version_ihl >> 4
	ihl = version_ihl & 0xF

	iph_length = ihl * 4

	ttl = iph[5]
	protocol = iph[6]
	s_addr = socket.inet_ntoa(iph[8])
	d_addr = socket.inet_ntoa(iph[9])

	# print 'Version : ' + str(version) + ' IP Header Length : ' + str(ihl) + ' TTL : ' + str(ttl) + ' Protocol : ' + str(protocol) + ' Source Address : ' + str(s_addr) + ' Destination Address : ' + str(d_addr)

	tcp_header = packet[iph_length:iph_length+20]

	#now unpack them :)
	tcph = unpack('!HHLLBBHHH' , tcp_header)

	source_port = tcph[0]
	dest_port = tcph[1]
	sequence = tcph[2]
	acknowledgement = tcph[3]
	doff_reserved = tcph[4]

	flags = tcph[5]

	syn = flags & 0b10
	ack = flags & 0b10000
	fin = flags & 0b1


	# If start connection
	if syn and not ack:
		# Show connection and port
		print(s_addr, dest_port)


	tcph_length = doff_reserved >> 4
	# print(acknowledgement)
	# print 'Source Port : ' + str(source_port) + ' Dest Port : ' + str(dest_port) + ' Sequence Number : ' + str(sequence) + ' Acknowledgement : ' + str(acknowledgement) + ' TCP header length : ' + str(tcph_length)

	h_size = iph_length + tcph_length * 4
	data_size = len(packet) - h_size

	#get data from the packet
	data = packet[h_size:]

	# print 'Data : ' + data
	# print



import pyqrcode
import os

OTPAUTH_URL = "otpauth://totp/{hostname}?secret={secret}&issuer={issuer}"


class OTPBidi():

    def __init__(self, secret):
        self._secret = secret.replace(' ', '')

    def generate(self):

        code = OTPAUTH_URL.format(
            **dict(
                hostname=os.uname().nodename,
                secret=self._secret,
                issuer='c-lock'
            )
        )

        qr = pyqrcode.create(code, error='L')
        return qr.terminal(quiet_zone=1)



import threading
import uuid
import logging

log = logging.getLogger(__name__)

def bypass(fa, fb):

    def mix(*args, **kwargs):
        fa(*args, **kwargs)
        fb(*args, **kwargs)

    return mix

class ProcWorker(threading.Thread):

    def __init__(self, i_q, o_q):
        super(ProcWorker, self).__init__()
        self._id = uuid.uuid4()
        print(self._id)
        self._i = i_q
        self._o = o_q
        self._end_evt = threading.Event()

    def run(self):
        while not self._end_evt.is_set():
            evt = self._i.get(True)
            if evt:
                self.process_evt(evt)


    def process_evt(self, evt):
        # TODO Send received event!
        if evt.get_id() == ProcWorkerEvent.END:
            log.debug("Received close signal")
            self._end_evt.set()
            self._i.put(None)


class Broker(ProcWorker):

    _client_qs = []

    def __init__(self, i_q, o_q):
        super(Broker, self).__init__(i_q, o_q)
        self._client_qs.append(o_q)

    def add_client(self, client_q):
        self._client_qs.append(client_q)

    def process_evt(self, evt):
        super(Broker, self).process_evt(evt)

        for q in self._client_qs:
            q.put(evt)


class Event():

    def __init__(self, id, value):
        self._id = id
        self._value = value

    def get_value(self):
        return self._value

    def get_id(self):
        return self._id

    def __str__(self):
        return "EVENT %s - %s" % (self._id, self._value )

#-----------------------#

class ProcWorkerEvent():

    NONE = -1
    END = uuid.uuid4().bytes


class PortManagerEvent():

    NEW_CONNECTION = uuid.uuid4().bytes
    CLOSING_SOCKET = uuid.uuid4().bytes
    FIRST_PORT = uuid.uuid4().bytes
    LAST_PORT = uuid.uuid4().bytes
    ERROR_OPENING_SOCKET = uuid.uuid4().bytes
    PROTECT_PORT = uuid.uuid4().bytes

class TocTocPortsEvent():

    NEW_SLOT = uuid.uuid4().bytes
    LAST_PORT = uuid.uuid4().bytes
    PROTECT_PORT = uuid.uuid4().bytes


class FirewallManagerEvent():

    NEW_CONNECTION = uuid.uuid4().bytes



import clock.totp as totp
import codecs

K = 0x3132333435363738393031323334353637383930
S = codecs.decode("%x" % K, "hex")


# OTP TEST
def test_otp():
    assert totp.otp(S, 0) == 1284755224
    assert totp.otp(S, 1) == 1094287082
    assert totp.otp(S, 2) == 137359152
    assert totp.otp(S, 3) == 1726969429
    assert totp.otp(S, 4) == 1640338314
    assert totp.otp(S, 5) == 868254676
    assert totp.otp(S, 6) == 1918287922
    assert totp.otp(S, 7) == 82162583
    assert totp.otp(S, 8) == 673399871
    assert totp.otp(S, 9) == 645520489


# HOTP TEST
def test_hotp():
    assert totp.hotp(S, 0) == 755224
    assert totp.hotp(S, 1) == 287082
    assert totp.hotp(S, 2) == 359152
    assert totp.hotp(S, 3) == 969429
    assert totp.hotp(S, 4) == 338314
    assert totp.hotp(S, 5) == 254676
    assert totp.hotp(S, 6) == 287922
    assert totp.hotp(S, 7) == 162583
    assert totp.hotp(S, 8) == 399871
    assert totp.hotp(S, 9) == 520489


# TOTP TEST
def test_totp():
    assert totp.totp(S, 59, n=8) == 94287082
    assert totp.totp(S, 1111111109, n=8) == 7081804
    assert totp.totp(S, 1234567890, n=8) == 89005924



from .proc_worker import ProcWorker, ProcWorkerEvent, TocTocPortsEvent
import uuid
import logging
import threading
import time
import iptc


log = logging.getLogger(__name__)


# GUIDE: https://github.com/ldx/python-iptables
class FirewallManager():

    def __init__(self):

        self.backup()

        log.debug("Starting FirewallManager")

        table = iptc.Table(iptc.Table.FILTER)

        # Crear chain
        try:
            table.create_chain("c-lock")
        except Exception as e:
            log.debug("c-lock exists!")


        # TODO ¿Debería venir desde ACCEPT?
        chain = iptc.Chain(iptc.Table(iptc.Table.FILTER), "INPUT")
        # TODO Añadir que mande aquí todos los puertos protegidos, o todas las conexiones si se protege todo
        rule = iptc.Rule() # *
        rule.protocol = "tcp"
        # Apuntar INPUT a c-lock
        rule.target = iptc.Target(rule, "c-lock")
        chain.insert_rule(rule, position=len(chain.rules))


        # c-lock config
        chain = iptc.Chain(iptc.Table(iptc.Table.FILTER), "c-lock")

        '''
        TODO 1b53c7b5-55d7-4834-9719-1ef86a7bfe12
        if unmanaged_ports:
            OPEN(unmanaged_ports)
            DROP_ALL
        else:
            DROP (PROTECTED_PORTS)
        '''
        # Drop all the rest
        rule = iptc.Rule()
        rule.protocol = "tcp"
        rule.target = iptc.Target(rule, "DROP")
        chain.insert_rule(rule)

        # Accept all established
        rule = iptc.Rule()
        rule.protocol = "tcp"
        rule.target = iptc.Target(rule, "ACCEPT")
        match = iptc.Match(rule, "state")
        match.state = "RELATED,ESTABLISHED"
        rule.add_match(match)
        chain.insert_rule(rule)

        # TODO Accept all OUTPUT

        # Accept all localhost connections
        rule = iptc.Rule() # *
        rule.protocol = "tcp"
        rule.src = "127.0.0.1"
        rule.target = iptc.Target(rule, "ACCEPT")
        chain.insert_rule(rule)

        # TODO Not working right
        # Accept all output connections
        # rule = iptc.Rule()
        # rule.protocol = "tcp"
        # rule.target = iptc.Target(rule, "ACCEPT")
        # rule.src = "127.0.0.1"
        # chain.insert_rule(rule)
    #
    # def unmanage_port(self, port):
    #
    #     table = iptc.Table(iptc.Table.FILTER)
    #
    #     chain = iptc.Chain(table, "c-lock-unmanaged")
    #
    #     rule = iptc.Rule() # *
    #     rule.protocol = "tcp"
    #     match = iptc.Match(rule, "tcp")
    #     match.dport = "%d" % port
    #     rule.add_match(match)
    #
    #     # TODO Debería ir a INPUT, pero puede hacer un bucle infinito
    #     rule.target = iptc.Target(rule, "ACCEPT")
    #
    #     chain.insert_rule(rule)


    # if !open then close
    def gen_rule(self, d_port=None, s_address=None, open=True):

        rule = iptc.Rule() # *
        rule.protocol = "tcp"

        if s_address:
            rule.src = s_address

        if d_port:
            print(d_port)
            match = iptc.Match(rule, "tcp")
            match.dport = "%d" % d_port
            rule.add_match(match)

        rule.target = iptc.Target(rule, "ACCEPT" if open else "REJECT")

        # TODO Puede servir para evitar repetidos
        # try:
        #     self.delete_rule(rule)
        # except Exception as e:
        #     pass

        return rule

    def open(self, d_port=None, s_address=None):
        # TODO Evitar insertar reglas repetidas
        table = iptc.Table(iptc.Table.FILTER)

        chain = iptc.Chain(table, "c-lock")

        rule = self.gen_rule(d_port, s_address, open=True)

        chain.insert_rule(rule)

        return rule

    def close(self, d_port=None, s_address=None):
        table = iptc.Table(iptc.Table.FILTER)

        chain = iptc.Chain(table, "c-lock")

        rule = self.gen_rule(d_port, s_address, open=False)

        chain.insert_rule(rule)

        return rule

    def add_rule(self, rule):

        table = iptc.Table(iptc.Table.FILTER)

        chain = iptc.Chain(table, "c-lock")
        chain.insert_rule(rule)

    def delete_rule(self, rule):
        table = iptc.Table(iptc.Table.FILTER)

        chain = iptc.Chain(table, "c-lock")
        chain.delete_rule(rule)

    def clean_firewall(self):
        log.info("Cleaning firewall rules")

        table = iptc.Table(iptc.Table.FILTER)

        chain = iptc.Chain(iptc.Table(iptc.Table.FILTER), "INPUT")
        rule = iptc.Rule()
        rule.protocol = "tcp"
        rule.target = iptc.Target(rule, "c-lock")

        # TODO Ver como usar esto sin borrar otras reglas del firewall
        while rule in chain.rules:
            chain.delete_rule(rule)

        chain = iptc.Chain(table, "c-lock")
        chain.flush()
        chain.delete()

    def finish(self):
        self.clean_firewall()
        self.restore()

    def backup(self):
        # TODO Search how to backup iptables
        pass

    def restore(self):
        # TODO Search how to restore iptables
        pass


class RuleManager(threading.Thread):

    rules = {}

    def __init__(self, fwm):
        super(RuleManager, self).__init__()
        self._lock = threading.Lock()
        self._fwm = fwm
        self._end_evt = threading.Event()
        self.start()

    def run(self):
        while not self._end_evt.is_set():
            self.delete_caduced_rules()
            self._end_evt.wait(1)

    def close(self):
        self._end_evt.set()
        self.delete_all_rules()

    def lock(f):

        def locker(self, *args, **kwargs):

            self._lock.acquire()

            res = f(self, *args, **kwargs)

            self._lock.release()

            return res

        return locker

    @lock
    def add_rule(self, r, caducity=-1, protected=False):

        # TODO rule_id como hash de la regla para evitar repetidos
        # Sin repetidos, sólo actualizaríamos la caducidad
        rule_id = str(uuid.uuid4())
        log.debug("Adding rule %s -> %s" % (rule_id, str(r)))

        # TODO Comprobar si ya existe
        self.rules[rule_id] = {
            'rule': r,
            'timestamp': time.time(),
            'caducity': caducity,
            'protected': protected
        }

        return rule_id

    @lock
    def exist_rule(self, r):

        for rule in self.rules:
            raux = self.rules[rule]
            if raux['rule'] == r:
                return rule

        return None

    @lock
    def renew_rule_timestamp(self, rule_id, caducity=None):

        if caducity:
            self.rules[rule_id]['caducity'] = caducity

        self.rules[rule_id]['timestamp'] = time.time()

    @lock
    def get_rule(self, rule_id):

        rule_data = self.rules.get(rule_id, None)

        return rule_data

    @lock
    def delete_rule(self, rule_id):

        if rule_id in self.rules:
            rule_data = self.rules[rule_id]
            try:
                log.debug("Deleting rule %s -> %s" % (rule_id, str(rule_data.get('rule'))))
                self._fwm.delete_rule(rule_data.get('rule'))
            except Exception as e:
                log.error("Error deleting %s: %s" % (rule_id, str(e)))

            del self.rules[rule_id]

    def delete_caduced_rules(self):
        # Copy of keys
        keys = [key for key in self.rules.keys()]

        for rule_id in keys:
            rule_data = self.get_rule(rule_id)
            if rule_data:
                if rule_data['caducity'] < 0:
                    continue
                elif rule_data['caducity'] < ((time.time() - rule_data['timestamp'])):
                    self.delete_rule(rule_id)

    # If `hard`, the protected rules are deleted too
    def delete_all_rules(self, hard=False):

        keys = [key for key in self.rules.keys()]

        for rule_id in keys:
            rule_data = self.get_rule(rule_id)
            if rule_data:
                # Delete if its not protected or hard deleting
                if hard or not rule_data['protected']:
                    self.delete_rule(rule_id)


class FirewallManagerWorker(ProcWorker):

    def __init__(self, i_q, o_q, fwm=None):

        super(FirewallManagerWorker, self).__init__(i_q, o_q)

        if not fwm:
            fwm = FirewallManager()

        self._fwm = fwm

        # TODO Abrir puertos marcados como no gestionados (prohibidos)
        self._rule_manager = RuleManager(fwm)

    def open(self, port=None, s_address=None, caducity=-1, protected=False):

        # We protect this rule for allowing the user to connect on step change
        r = self._fwm.gen_rule(port, s_address=s_address)

        exist = self._rule_manager.exist_rule(r)

        if exist:
            self._rule_manager.renew_rule_timestamp(exist)
        else:
            try:
                self._fwm.add_rule(r)
                self._rule_manager.add_rule(r, caducity=caducity, protected=protected)
                log.debug("Opening port %d for %s" % (port, s_address))
            except Exception as e:
                log.critical("Error opening port for %s %s" % (s_address, e))

    def process_evt(self, evt):

        super(FirewallManagerWorker, self).process_evt(evt)

        if evt.get_id() == ProcWorkerEvent.END:
            self._rule_manager.close()
            self._fwm.finish()

        if evt.get_id() == TocTocPortsEvent.LAST_PORT:
            evt_value = evt.get_value()
            ports = evt_value['ports']
            addr = evt_value['address']
            log.info("Opening ports %s for %s" % (ports, addr))

            # Usamos esto porque hemos determinado usar FT/SNIFF/O1
            if len(ports):
                for port in ports:
                    self.open(port, s_address=addr, caducity=30, protected=True)
            else:
                # TODO Implementar esto si determinamos usar FT/SNIFF/O2
                self.open(s_address=addr, caducity=30, protected=True)

        if evt.get_id() == TocTocPortsEvent.NEW_SLOT:
            # TODO ¿Close o borrar las reglas guardadas?
            # self._fwm.close()
            self._rule_manager.delete_all_rules()



from setuptools import setup, find_packages


def requirements():
    with open('requirements.txt') as f:
        raw = f.read()

    return [dependency for dependency in raw.split('\n') if len(dependency)]


with open('README.md') as f:
    readme = f.read()

setup(
    name="c-lock",
    version="0.0.7.1",
    description="",
    long_description=readme,
    long_description_content_type='text/markdown',
    python_requires=">=3.6",
    author="Javier Junquera Sánchez",
    author_email="javier@junquera.xyz",
    # packages=find_packages(),
    packages=["clock"],
    url="https://gitlab.com/junquera/c-lock",
    license="MIT License",
    install_requires=requirements(),
    tests_require=["pytest"],
    entry_points=dict(
        console_scripts= [
            'c-lockd=clock.cli.server:main',
            'c-lock=clock.cli.client:main'
        ]
    )
)



import threading
from queue import Queue
import time

class Proc():

    def __init__(self):
        pass

    def a(self):
        print("A")

    def b(self):
        print("B")


class ProcWorker(threading.Thread):

    def __init__(self, i_q, o_q):
        super(ProcWorker, self).__init__()
        self.instance = Proc()
        self.i = i_q
        self.o = o_q

    def run(self):

        while 1:
            aob = self.i.get(True)
            time.sleep(1)
            if aob == 'a':
                self.instance.a()
                self.o.put('b')
            else:
                self.instance.b()
                self.o.put('a')

q1 = Queue()
q2 = Queue()

ProcWorker(q1, q2).start()
ProcWorker(q2, q1).start()
print("All running")

q2.put('b')



import socket, sys
from struct import *
import threading
from .proc_worker import ProcWorker, Event, bypass, ProcWorkerEvent, TocTocPortsEvent, PortManagerEvent

import logging

log = logging.getLogger(__name__)

from scapy.all import sniff, IP, TCP

class PortManager():

    def __init__(self, address='0.0.0.0', unmanaged_ports=[]):

        self._sockets = []
        self._threads = []
        self._address = address
        self._unmanaged_ports = unmanaged_ports

        try:

            evt = threading.Event()
            t = threading.Thread(target=self.wait_and_listen, args=(evt,))
            self._threads.append(evt)
            t.start()

        except socket.error:
            # TODO Send END
            print( 'Socket could not be created. Error Code : ' + str(msg[0]) + ' Message ' + msg[1])
            sys.exit()

    def wait_and_listen(self, evt):
        log.info("wait_and_listen")

        # TODO Ver por qué no termina el hilo...
        myfilter = '(tcp[13]&2!=0 and tcp[13]&16==0)'
        sniff(prn=lambda pkt: self.notify_connection(pkt[IP].src, pkt[TCP].dport), stop_filter=lambda x: evt.is_set(), filter=myfilter, store=0)

        log.info("nor_wait_nor_listen")

    def notify_connection(self, addr, port):
        # print(addr, port)
        # TODO Hacer esto con métodos con bloqueos (@lock)
        if addr in self._active:
            addr_info = self._active[addr]
            if port == addr_info['next']:
                next_n = addr_info['n'] + 1
                if len(self._port_list) <= next_n:
                    self.last_port(addr)
                    del self._active[addr]
                else:
                    addr_info['n'] = next_n
                    addr_info['next'] = self._port_list[next_n]
                    self._active[addr] = addr_info
            else:
                if port not in self._unmanaged_ports:
                    del self._active[addr]
        else:
            if self._port_list[0] == port:
                self._active[addr] = dict(next=self._port_list[1], n=1)

    def last_port(self, addr):
        log.info("%s reached last port" % (addr))

    # TODO Rename to reset
    def open(self, port_list):
        self._active = {}
        self._port_list = port_list

    def close_thread(self, evt):
        try:
            evt.set()
        except Exception as e:
            pass

    def unlock_threads(self):
        while len(self._threads):
            try:
                evt = self._threads.pop()
                self.close_thread(evt)
            except Exception as e:
                pass

    def close(self):
        self.unlock_threads()


# https://eli.thegreenplace.net/2011/12/27/python-threads-communication-and-stopping
# http://www.bogotobogo.com/python/Multithread/python_multithreading_Event_Objects_between_Threads.php
class PortManagerWorker(ProcWorker):

    def __init__(self, i_q, o_q, pm=None):
        super(PortManagerWorker, self).__init__(i_q, o_q)

        if not pm:
            pm = PortManager()

        self._pm = pm

        self._pm.notify_connection = bypass(self._pm.notify_connection, self.notify_connection)
        self._pm.last_port = bypass(self._pm.last_port, self.last_port)

    def notify_connection(self, addr, port):
        self._o.put(Event(PortManagerEvent.NEW_CONNECTION, {'port': port, 'address': addr}))

    def last_port(self, address):
            self._o.put(Event(PortManagerEvent.LAST_PORT, dict(address=address)))

    def process_evt(self, evt):
        super(PortManagerWorker, self).process_evt(evt)

        if evt.get_id() == ProcWorkerEvent.END:
            self._pm.close()

        if evt.get_id() == TocTocPortsEvent.NEW_SLOT:
            port_list = evt.get_value()['port_list'].get_values()
            self._pm.open(port_list)

        if evt.get_id() == PortManagerEvent.PROTECT_PORT:
            pass
